/*
 * Copyright (c) 2022, Oracle and/or its affiliates.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2.0, as
 * published by the Free Software Foundation.
 *
 * This program is also distributed with certain software (including
 * but not limited to OpenSSL) that is licensed under separate terms,
 * as designated in a particular file or component or in included license
 * documentation.  The authors of MySQL hereby grant you an
 * additional permission to link the program and your derivative works
 * with the separately licensed software that they have included with
 * MySQL.
 *
 * Without limiting anything contained in the foregoing, this file,
 * which is part of MySQL Connector/Node.js, is also subject to the
 * Universal FOSS Exception, version 1.0, a copy of which can be found at
 * http://oss.oracle.com/licenses/universal-foss-exception.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License, version 2.0, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
 */

import { Any } from '../Protocol/Datatypes';
import Result from './Result';

/**
 * A plain JavaScript object or JSON literal representing an X DevAPI document
 * As defined by the [DocumentOrJSON](https://dev.mysql.com/doc/x-devapi-userguide/en/crud-ebnf-other-definitions.html#crud-ebnf-documentorjson)
 * EBNF definition.
 */
export type DocumentOrJSON = { [key: string]: Any } | string;

/**
 * One ore more DocumentOrJSON instances.
 */
export type CollectionDocuments = DocumentOrJSON | DocumentOrJSON[];

/**
 * Statement generated by `collection.add()`.
 */
interface CollectionAdd {
    /**
     * Adds a new document to the collection.
     * @param doc - One or more documents or arrays of documents (plain JavaScript objects or JSON literals).
     * @returns The current statement instance itself.
     */
    add: (doc: CollectionDocuments, ...docs: CollectionDocuments[]) => CollectionAdd
    /**
     * Executes the current statement.
     * @returns A `Promise` that resolves to a `DocResult` instance once the statement is executed.
     */
    execute: () => Promise<Result>
}

export default CollectionAdd;
