/*
 * Copyright (c) 2022, Oracle and/or its affiliates.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2.0, as
 * published by the Free Software Foundation.
 *
 * This program is also distributed with certain software (including
 * but not limited to OpenSSL) that is licensed under separate terms,
 * as designated in a particular file or component or in included license
 * documentation.  The authors of MySQL hereby grant you an
 * additional permission to link the program and your derivative works
 * with the separately licensed software that they have included with
 * MySQL.
 *
 * Without limiting anything contained in the foregoing, this file,
 * which is part of MySQL Connector/Node.js, is also subject to the
 * Universal FOSS Exception, version 1.0, a copy of which can be found at
 * http://oss.oracle.com/licenses/universal-foss-exception.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License, version 2.0, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
 */

import { Expr } from '../Protocol/Expr';
import Binding from './Binding';
import Ordering from './CollectionOrdering';
import DocResult from './DocResult';
import Grouping from './Grouping';
import Limiting from './Limiting';
import Locking from './Locking';
import Skipping from './Skipping';

/**
 * A JavaScript string or a protobuf expression object.
 * As defined by the [ProjectedDocumentExprStr](https://dev.mysql.com/doc/x-devapi-userguide/en/crud-ebnf-other-definitions.html#crud-ebnf-projecteddocumentexprstr)
 * EBNF definition.
 */
export type ProjectedDocumentExprStr = string | Expr;

/**
 * One or more ProjectedDocumentExprStr instances.
 */
export type ProjectedDocumentExprStrList = ProjectedDocumentExprStr | ProjectedDocumentExprStr[];

/**
 * Statement generated by `collection.find()`.
 */
interface CollectionFind extends Binding<CollectionFind>, Grouping<CollectionFind>, Limiting<CollectionFind>, Locking<CollectionFind>, Ordering<CollectionFind>, Skipping<CollectionFind> {
    /**
     * Executes the current statement.
     * @returns A `Promise` that resolves to a `DocResult` instance once the statement is executed.
     */
    execute: () => Promise<DocResult>
    /**
     * Picks specific document fields to include in the result set.
     * @param expr - One or more strings, X Protocol message objects, arrays
     * of strings or X Protocol message objects encoding the X DevAPI
     * expression.
     * @param exprs - Additional strings, X Protocol message objects, arrays
     * @returns The current statement instance itself.
     */
    fields: (expr: ProjectedDocumentExprStrList, ...exprs: ProjectedDocumentExprStrList[]) => CollectionFind
}

export default CollectionFind;
